'' =================================================================================================
''
''   File....... jm_mono_music.spin
''   Purpose....
''   Author..... Jon McPhalen
''               Copyright (c) 2018-2024 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 06 DEC 2024
''
'' =================================================================================================


con { timing }

  _xinfreq = 5_000_000                                          ' use 5MHz crystal
  _clkmode = xtal1 + pll16x                                     ' x16 = 80MHz

  CLK_FREQ = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us


con { terminal }

  BR_TERM  = 115_200                                            ' terminal baud rate


con { fixed io pins }

  PGM_RX   = 31  { I }                                          ' serial / programming
  PGM_TX   = 30  { O }

  EE_SDA   = 29  { I/O }                                        ' i2c / eeprom
  EE_SCL   = 28  { I/O }


con { app io pins }

  SPEAKER  =  0  { O }


con

  #true,  ON, OFF
  #false, NO, YES

  DEBUG_ON = NO


obj

' main                                                          ' * master Spin cog
  time : "jm_time_80"                                           '   timing and delays (80MHz)
  prng : "jm_prng"                                              '   pseudo-random number generation
  io   : "jm_io"                                                '   simple io
  term : "jm_fullduplexserial"                                  ' * serial IO for terminal

' * uses cog when loaded


dat

  Charge        long    N_G4, T_8TH,         N_C5, T_8TH, N_E5, T_8TH
                long    N_G5, T_8TH * 3 / 2, N_E5, T_8TH, N_G5, T_DHALF
  EndOfCharge   long    0, 0, -1, -1


var { globals }


pub main

  setup
  wait_for_terminal(true, 250)

  start_music(@Charge, BPM)                                     ' play at standard rate

  repeat
    waitcnt(0)


pub setup

'' Setup IO and objects for application

  time.start                                                    ' setup timing & delays

  prng.seed(-cnt, -cnt ~> 2, $CAFE_BABE, cnt <- 2, cnt)         ' seed randomizer

  io.start(0, 0)                                                ' clear all pins (master cog)

  bcog := cognew(background, @bcstack) + 1                      ' start background cog *

  if (DEBUG_ON == YES)
    term.tstart(BR_TERM)                                        ' start serial for terminal *


pub wait_for_terminal(clear, delay)

'' Wait for terminal to be open and key pressed

  if (DEBUG_ON == YES)
    term.rxflush
    term.rx
    if (clear)
      term.tx(term#CLS)
      time.pause(delay #> 0)


con { bkg support }

  { ----------------------------------------- }
  {  B / G   S U P P O R T   R O U T I N E S  }
  { ----------------------------------------- }


pub start_tone(freq, ms)

'' Play tone of freq for ms milliseconds
'' -- non-blocking
'' -- freq in 0.1Hz units

  if (p_music > 0)                                              ' busy
    longfill(@p_music, 0, 2)                                    ' kill music
    waitcnt(cnt + MS_001)                                       ' - give it time to die

  longmove(@OneNote, @freq, 2)                                  ' setup note

  BPM := 60                                                     ' play it
  p_music := @OneNote


pub stop_tone

'' Kills background note if playing

  longfill(@p_music, 0, 2)


pub start_music(p_notes, rate)

'' Play music (notes/timing) at p_notes
'' -- non-blocking
'' -- rate is BPM

  BPM     := rate
  p_music := p_notes


pub stop_music

'' Kills background music if playing

  longfill(@p_music, 0, 2)


pub playing

'' Returns true when background tone/music is playing

  return (p_music > 0)


con { background }

  { ----------------------------- }
  {  B A C K G R O U N D   C O G  }
  {  - global milliseconds        }
  {  - play monophonic music      }
  {  - play simple tone           }
  { ----------------------------- }


var

  long  bcog                                                    ' background cog #
  long  bcstack[64]                                             ' stack for background cog

  long  millis                                                  ' global milliseconds register


pri background | t                                              ' launch with cognew()

  millis := 0

  t := cnt                                                      ' sync loop timer
  repeat
    waitcnt(t += MS_001)                                        ' run loop every millisecond
    ++millis
    play_music


var

  long  p_music
  long  nmillis


dat

  BPM           long    120

  OneNote       long    0, 0, -1, -1                            ' for tone_on


pri play_music | nfreq

  if (p_music =< 0)                                             ' manual stop invoked?
    io.freq_out(0, SPEAKER, 0)
    return

  if (--nmillis > 0)                                            ' is note playing now?
    return

  nfreq := long[p_music]                                        ' get note
  p_music += 4                                                  ' advance to timing
  if (nfreq < 0)                                                ' if end
    longfill(@p_music, 0, 2)
    return

  nmillis := long[p_music] * 60 / BPM                           ' adjust timing for bpm
  p_music += 4

  io.freq_out(0, SPEAKER, nfreq)                                ' start the note


con { musical notes }

  ' base timing at 60 BPM

  T_WHOLE = 4000
  T_DHALF = T_WHOLE * 3 / 4
  T_HALF  = T_WHOLE >> 1
  T_4TH   = T_WHOLE >> 2
  T_8TH   = T_WHOLE >> 3
  T_16TH  = T_WHOLE >> 4

  ' frequencies in 0.1Hz units

  N_C0    =   16_3
  N_Cs0   =   17_3                                             ' sharp
  N_D0    =   18_3
  N_Ds0   =   19_4
  N_E0    =   20_6
  N_F0    =   21_8
  N_Fs0   =   23_1
  N_G0    =   24_5
  N_Gs0   =   25_9
  N_A0    =   27_5
  N_As0   =   29_1
  N_B0    =   30_8

  N_C1    =   32_7
  N_Cs1   =   34_6
  N_D1    =   36_7
  N_Ds1   =   38_8
  N_E1    =   41_2
  N_F1    =   43_6
  N_Fs1   =   46_2
  N_G1    =   49_0
  N_Gs1   =   51_9
  N_A1    =   55_0
  N_As1   =   58_2
  N_B1    =   61_7

  N_C2    =   65_4
  N_Cs2   =   69_3
  N_D2    =   73_4
  N_Ds2   =   77_8
  N_E2    =   82_4
  N_F2    =   87_3
  N_Fs2   =   92_5
  N_G2    =   98_0
  N_Gs2   =  103_8
  N_A2    =  110_0
  N_As2   =  116_5
  N_B2    =  123_5

  N_C3    =  130_8
  N_Cs3   =  138_6
  N_D3    =  146_8
  N_Ds3   =  155_6
  N_E3    =  164_8
  N_F3    =  174_6
  N_Fs3   =  185_0
  N_G3    =  196_0
  N_Gs3   =  207_7
  N_A3    =  220_0
  N_As3   =  233_0
  N_B3    =  247_0

  N_C4    =  261_6
  N_Cs4   =  277_2
  N_D4    =  293_7
  N_Ds4   =  311_1
  N_E4    =  329_6
  N_F4    =  349_2
  N_Fs4   =  370_0
  N_G4    =  392_0
  N_Gs4   =  415_3
  N_A4    =  440_0
  N_As4   =  466_1
  N_B4    =  493_9

  N_C5    =  523_3
  N_Cs5   =  554_4
  N_D5    =  587_3
  N_Ds5   =  622_2
  N_E5    =  659_2
  N_F5    =  698_5
  N_Fs5   =  740_0
  N_G5    =  784_0
  N_Gs5   =  830_6
  N_A5    =  880_0
  N_As5   =  932_3
  N_B5    =  987_8

  N_C6    = 1046_5
  N_Cs6   = 1108_7
  N_D6    = 1174_7
  N_Ds6   = 1244_5
  N_E6    = 1318_5
  N_F6    = 1396_9
  N_Fs6   = 1480_0
  N_G6    = 1568_0
  N_Gs6   = 1661_2
  N_A6    = 1760_0
  N_As6   = 1864_7
  N_B6    = 1975_5

  N_C7    = 2093_0
  N_Cs7   = 2217_5
  N_D7    = 2349_3
  N_Ds7   = 2489_0
  N_E7    = 2637_0
  N_F7    = 2793_8
  N_Fs7   = 2960_0
  N_G7    = 3136_0
  N_Gs7   = 3322_4
  N_A7    = 3520_0
  N_As7   = 3729_3
  N_B7    = 3951_0

  N_C8    = 4186_0
  N_Cs8   = 4434_9
  N_D8    = 4698_6
  N_Ds8   = 4978_0
  N_E8    = 5274_0
  N_F8    = 5587_7
  N_Fs8   = 5919_9
  N_G8    = 6271_9
  N_Gs8   = 6644_9
  N_A8    = 7040_0
  N_As8   = 7458_6
  N_B8    = 7902_1


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}